/// <reference path='../node_modules/typescript/bin/typescriptServices.d.ts' />
/// <reference path='../typings/tsd.d.ts' />
var Promise = require('bluebird');
var _ = require('lodash');
var loaderUtils = require('loader-utils');
var host = require('./host');
var deps = require('./deps');
var helpers = require('./helpers');
function getRootCompiler(compiler) {
    if (compiler.parentCompilation) {
        return getRootCompiler(compiler.parentCompilation.compiler);
    }
    else {
        return compiler;
    }
}
function getInstanceStore(compiler) {
    var store = getRootCompiler(compiler)._tsInstances;
    if (store) {
        return store;
    }
    else {
        throw new Error('Can not resolve instance store');
    }
}
function ensureInstanceStore(compiler) {
    var rootCompiler = getRootCompiler(compiler);
    if (!rootCompiler._tsInstances) {
        rootCompiler._tsInstances = {};
    }
}
function resolveInstance(compiler, instanceName) {
    return getInstanceStore(compiler)[instanceName];
}
function ensureInstance(webpack, options, instanceName) {
    ensureInstanceStore(webpack._compiler);
    var exInstance = resolveInstance(webpack._compiler, instanceName);
    if (exInstance) {
        return exInstance;
    }
    var tsFlow = Promise.resolve();
    var tsImpl;
    if (options.compiler) {
        tsImpl = require(options.compiler);
    }
    else {
        tsImpl = require('typescript');
    }
    var configFileName = tsImpl.findConfigFile(options.tsconfig || process.cwd());
    var configFile = null;
    if (configFileName) {
        configFile = tsImpl.readConfigFile(configFileName);
        if (configFile.error) {
            throw configFile.error;
        }
        _.extend(options, configFile.compilerOptions);
        _.extend(options, configFile.awesomeTypescriptLoaderOptions);
    }
    if (typeof options.emitRequireType === 'undefined') {
        options.emitRequireType = true;
    }
    else {
        if (typeof options.emitRequireType === 'string') {
            options.emitRequireType = options.emitRequireType === 'true';
        }
    }
    if (typeof options.reEmitDependentFiles === 'undefined') {
        options.reEmitDependentFiles = false;
    }
    else {
        if (typeof options.reEmitDependentFiles === 'string') {
            options.reEmitDependentFiles = options.reEmitDependentFiles === 'true';
        }
    }
    if (typeof options.useWebpackText === 'undefined') {
        options.useWebpackText = false;
    }
    else {
        if (typeof options.useWebpackText === 'string') {
            options.useWebpackText = options.useWebpackText === 'true';
        }
    }
    if (typeof options.rewriteImports == 'undefined') {
        options.rewriteImports = '';
    }
    if (options.target) {
        options.target = helpers.parseOptionTarget(options.target, tsImpl);
    }
    var tsState = new host.State(options, webpack._compiler.inputFileSystem, tsImpl);
    var compiler = webpack._compiler;
    compiler.plugin('watch-run', function (watching, callback) {
        var resolver = Promise.promisify(watching.compiler.resolvers.normal.resolve);
        var instance = resolveInstance(watching.compiler, instanceName);
        var state = instance.tsState;
        var mtimes = watching.compiler.watchFileSystem.watcher.mtimes;
        var changedFiles = Object.keys(mtimes);
        changedFiles.forEach(function (changedFile) {
            state.fileAnalyzer.validFiles.markFileInvalid(changedFile);
        });
        Promise.all(changedFiles.map(function (changedFile) {
            if (/\.ts$|\.d\.ts$/.test(changedFile)) {
                return state.readFileAndUpdate(changedFile).then(function () {
                    return state.fileAnalyzer.checkDependencies(resolver, changedFile);
                });
            }
            else {
                return Promise.resolve();
            }
        }))
            .then(function (_) { state.updateProgram(); callback(); })
            .catch(function (err) { return console.error(err); });
    });
    compiler.plugin('after-compile', function (compilation, callback) {
        var instance = resolveInstance(compilation.compiler, instanceName);
        var state = instance.tsState;
        var diagnostics = state.ts.getPreEmitDiagnostics(state.program);
        var emitError = function (err) {
            compilation.errors.push(new Error(err));
        };
        var phantomImports = [];
        Object.keys(state.files).forEach(function (fileName) {
            if (!instance.compiledFiles[fileName]) {
                phantomImports.push(fileName);
            }
        });
        instance.compiledFiles = {};
        compilation.fileDependencies.push.apply(compilation.fileDependencies, phantomImports);
        compilation.fileDependencies = _.uniq(compilation.fileDependencies);
        var errors = helpers.formatErrors(diagnostics);
        errors.forEach(emitError);
        callback();
    });
    return getInstanceStore(webpack._compiler)[instanceName] = {
        tsFlow: tsFlow,
        tsState: tsState,
        compiledFiles: {},
        options: options
    };
}
function loader(text) {
    compiler.call(undefined, this, text);
}
function compiler(webpack, text) {
    if (webpack.cacheable) {
        webpack.cacheable();
    }
    var options = loaderUtils.parseQuery(webpack.query);
    var instanceName = options.instanceName || 'default';
    var instance = ensureInstance(webpack, options, instanceName);
    var state = instance.tsState;
    var callback = webpack.async();
    var fileName = webpack.resourcePath;
    var resolver = Promise.promisify(webpack.resolve);
    var depsInjector = {
        add: function (depFileName) { webpack.addDependency(depFileName); },
        clear: webpack.clearDependencies.bind(webpack)
    };
    var applyDeps = _.once(function () {
        depsInjector.clear();
        depsInjector.add(fileName);
        if (state.options.reEmitDependentFiles) {
            state.fileAnalyzer.dependencies.applyChain(fileName, depsInjector);
        }
    });
    instance.tsFlow = instance.tsFlow
        .then(function () {
        instance.compiledFiles[fileName] = true;
        var doUpdate = false;
        if (instance.options.useWebpackText) {
            if (state.updateFile(fileName, text, true)) {
                doUpdate = true;
            }
        }
        return state.fileAnalyzer.checkDependencies(resolver, fileName).then(function (wasChanged) {
            if (doUpdate || wasChanged) {
                state.updateProgram();
            }
        });
    })
        .then(function () {
        return state.emit(fileName);
    })
        .then(function (output) {
        var result = helpers.findResultFor(output, fileName);
        if (result.text === undefined) {
            throw new Error('no output found for ' + fileName);
        }
        var sourceFilename = loaderUtils.getRemainingRequest(webpack);
        var current = loaderUtils.getCurrentRequest(webpack);
        var sourceMap = JSON.parse(result.sourceMap);
        sourceMap.sources = [sourceFilename];
        sourceMap.file = current;
        sourceMap.sourcesContent = [text];
        applyDeps();
        callback(null, result.text, sourceMap);
    })
        .finally(function () {
        applyDeps();
    })
        .catch(deps.ResolutionError, function (err) {
        console.error(err);
        callback(err, helpers.codegenErrorReport([err]));
    })
        .catch(function (err) { console.error(err); callback(err); });
}
module.exports = loader;
//# sourceMappingURL=index.js.map